/**
 * Copyright (c) 2009-2012 Zauber S.A. <http://www.zaubersoftware.com/>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.zauberlabs.commons.mom;


import java.util.List;
import java.util.Map;

import net.sf.staccatocommons.check.Ensure;
import net.sf.staccatocommons.collections.stream.Streams;
import net.sf.staccatocommons.lang.SoftException;
import net.sf.staccatocommons.lang.function.AbstractFunction;
import net.sf.staccatocommons.restrictions.check.NonNull;

/**
 * Abstract implementation of a {@link MapObjectMapper}. Subclasses need only to
 * define how to convert {@link Map} to Pojos and back, and implement coercions
 *
 * @author flbulgarelli
 */
public abstract class AbstractMapObjectMapper implements MapObjectMapper {

    /**
     * @see MapObjectMapper#unmap(java.lang.Object)
     */
    public Object unmap(final Object map) {
        return unmapImpl(map, null);
    }

    /**
     * @see MapObjectMapper#unmap(java.lang.Object, java.lang.Class)
     */
    public Object unmap(final Object value, @NonNull final Class<?> type) {
        return unmapImpl(value, type);
    }

    @SuppressWarnings("unchecked")
    protected Object unmapImpl(final Object value, final Class<?> type) {
        if (value == null) {
            return null;
        }
        if (value instanceof Map) {
            return unmapMap((Map) value, type);
        }
        if (value instanceof List) {
            return unmapList((List) value, type);
        }
        return coerce(value, type);
    }

    /**
     * Unmaps a {@link Map} object
     * @param map the map to convert into a bean
     * @param type the type of bean to convert to
     * @return the unmaped bean
     */
    protected final <A> A unmapMap(final Map<String, Object> map, final Class<A> type) {
        final Class<A> explicitType = explicitType(map);
        if (explicitType != null) {
            return unmapMapImpl(map, explicitType);
        }
        if (type != null) {
            return unmapMapImpl(map, type);
        }
        return Ensure.that().fail("Must specify key %s", JAVA_TYPE_KEY);

    }

    /**
     * Unmaps a {@link List} object
     * @param objects the list of object to convert into a bean
     * @param componentType the type of the objects in the list to convert to
     * @return the list of converted objects
     */
    protected final List<Object> unmapList(final List<Object> objects,
            final Class<?> componentType) {
        return Streams.from(objects).map(
                new AbstractFunction<Object, Object>() {
                    public Object apply(final Object arg0) {
                        return unmap(arg0, componentType);
                    }
                }).toList();
    }

    @SuppressWarnings("unchecked")
    protected <A> Class<A> explicitType(final Map<String, Object> map) {
        if(!map.containsKey(JAVA_TYPE_KEY)) {
            return null;
        }
        final Object javaType = map.get(JAVA_TYPE_KEY);
        Ensure.that() //
                .isNotNull(JAVA_TYPE_KEY, javaType) //
                .isInstanceOf(JAVA_TYPE_KEY, javaType, String.class);
        try {
            return (Class<A>) Class.forName((String) javaType);
        } catch (final ClassNotFoundException e) {
            throw SoftException.soften(e);
        }
    }

    protected abstract <A> A unmapMapImpl(Map<String, Object> map,
            Class<A> componentType);

    /**
     * Converts an object to the given class, if possible. Throws an exception, otherwise.
     *
     * @param value the instance to convert
     * @param type the type to convert to
     * @return the converted object
     */
    protected abstract Object coerce(Object value, Class<?> type);

    @SuppressWarnings("unchecked")
    public Object map(final Object value) {
        if (value == null) {
            return null;
        }
        if (value instanceof List) {
            return this.mapList((List) value);
        }
        if (this.isCompositeType(value.getClass())) {
            return this.mapObject(value);
        }
        return value;
    }

    protected final List<Object> mapList(final List<Object> values) {
        return Streams.from(values).map(
                new AbstractFunction<Object, Object>() {
                    public Object apply(final Object arg0) {
                        return map(arg0);
                    }
                }).toList();
    }

    protected final Map<String, Object> mapObject(final Object value) {
        return mapObjectImpl(value);
    }

    protected abstract Map<String, Object> mapObjectImpl(Object value);

    /**
     * Answers whether instances of the given type
     * should be converted into map when performing {@link #map(Object)}
     *
     * @param type the type to test
     * @return if type is a composite type
     */
    protected abstract boolean isCompositeType(@NonNull Class<?> type);

}
